@@ -1,10 +1,16 @@
+import copy
+from math import e
+import os
 import time
+import warnings
 
 import numpy as np
 import pandas as pd
 from IPython.display import display
+from joblib import dump, load
+from luciferml.supervised.utils.best import Best
 from luciferml.supervised.utils.classificationPredictor import classificationPredictor
-from luciferml.supervised.utils.configs import classifiers
+from luciferml.supervised.utils.configs import *
 from luciferml.supervised.utils.confusionMatrix import confusionMatrix
 from luciferml.supervised.utils.dimensionalityReduction import dimensionalityReduction
 from luciferml.supervised.utils.encoder import encoder
@@ -48,6 +54,8 @@ class Classification:
         k_neighbors=1,
         dropout_rate=0,
         verbose=False,
+        exclude_models=[],
+        path=None,
     ):
         """
         Encode Categorical Data then Applies SMOTE , Splits the features and labels in training and validation sets with test_size = .2 , scales self.X_train, self.X_val using StandardScaler.
@@ -110,15 +118,15 @@ class Classification:
                     No. of components for LDA. Default = 1
             n_components_pca : int
                     No. of components for PCA. Default = 2
-            self.hidden_layers : int
+            hidden_layers : int
                     No. of default layers of ann. Default = 4
             inputs_units : int
                     No. of units in input layer. Default = 6
             output_units : int
                     No. of units in output layer. Default = 6
-            self.input_activation : str
+            input_activation : str
                     Activation function for Hidden layers. Default = 'relu'
-            self.output_activation : str
+            output_activation : str
                     Activation function for Output layers. Default = 'sigmoid'
             optimizer: str
                     Optimizer for ann. Default = 'adam'
@@ -144,6 +152,11 @@ class Classification:
                 No. of neighbours for SMOTE. Default = 1
             verbose : boolean
                 Verbosity of models. Default = False
+            exclude_models : list
+                List of models to be excluded when using predictor = 'all' . Default = []
+            path : list
+                List containing path to saved model and scaler. Default = None
+                Example: [model.pkl, scaler.pkl]
 
         Returns:
             Dict Containing Name of Classifiers, Its K-Fold Cross Validated Accuracy and Prediction set
@@ -195,6 +208,7 @@ class Classification:
         self.smote = smote
         self.k_neighbors = k_neighbors
         self.verbose = verbose
+        self.exclude_models = exclude_models
 
         self.accuracy_scores = {}
         self.reg_result = {}
@@ -208,10 +222,25 @@ class Classification:
         self.acc = []
         self.bestacc = []
         self.bestparams = []
-        self.result_df = pd.DataFrame(index=classifiers.values())
+        self.best_classifier_path = ""
+        self.scaler_path = ""
+        self.classifier_model = []
+        self.result_df = pd.DataFrame(index=None)
+        self.classifiers = copy.deepcopy(classifiers)
+        for i in self.exclude_models:
+            self.classifiers.pop(i)
+        self.result_df["Name"] = list(self.classifiers.values())
+        self.best_classifier = "First Run the Predictor in All mode"
 
         self.pred_mode = ""
 
+        if path != None:
+            try:
+                self.classifier, self.sc = self.__load(path)
+            except Exception as e:
+                print(e)
+                print("Model not found")
+
     def fit(self, features, labels):
         """[Takes Features and Labels and Encodes Categorical Data then Applies SMOTE , Splits the features and labels in training and validation sets with test_size = .2
         scales X_train, self.X_val using StandardScaler.
@@ -282,7 +311,7 @@ class Classification:
         # Models ---------------------------------------------------------------------
         if self.predictor == "all":
             self.pred_mode = "all"
-            self._fitall()
+            self.__fitall()
             return
 
         elif self.predictor == "ann":
@@ -391,134 +420,153 @@ class Classification:
         self.end = time.time()
         print("Time Elapsed : ", self.end - self.start, "seconds \n")
 
-    def _fitall(self):
-        print("Training All Classifiers [*]")
-        for index, self.predictor in enumerate(classifiers):
-            if not self.predictor == "ann":
-                self.parameters, self.classifier = classificationPredictor(
-                    self.predictor,
-                    self.params,
-                    self.X_train,
-                    self.X_val,
-                    self.y_train,
-                    self.y_val,
-                    self.epochs,
-                    self.hidden_layers,
-                    self.input_activation,
-                    self.output_activation,
-                    self.loss,
-                    self.batch_size,
-                    self.metrics,
-                    self.validation_split,
-                    self.optimizer,
-                    self.output_units,
-                    self.input_units,
-                    self.tune_mode,
-                    all_mode=True,
-                    verbose=self.verbose,
-                )
-            elif self.predictor == "ann":
-                (
-                    self.parameters,
-                    self.classifier,
-                    self.classifier_wrap,
-                ) = classificationPredictor(
-                    self.predictor,
-                    self.params,
-                    self.X_train,
-                    self.X_val,
-                    self.y_train,
-                    self.y_val,
-                    self.epochs,
-                    self.hidden_layers,
-                    self.input_activation,
-                    self.output_activation,
-                    self.loss,
-                    self.batch_size,
-                    self.metrics,
-                    self.validation_split,
-                    self.optimizer,
-                    self.output_units,
-                    self.input_units,
-                    self.tune_mode,
-                    self.dropout_rate,
-                    all_mode=True,
-                    verbose=self.verbose,
-                )
-            try:
-
+    def __fitall(self):
+        print("Training All Classifiers [*]\n")
+        if self.params != {}:
+            warnings.warn(params_use_warning, UserWarning)
+            self.params = {}
+        for _, self.predictor in enumerate(self.classifiers):
+            if not self.predictor in self.exclude_models:
                 if not self.predictor == "ann":
-                    self.classifier.fit(self.X_train, self.y_train)
-            except Exception as error:
-                print(
-                    classifiers[self.predictor],
-                    "Model Train Failed with error: ",
-                    error,
-                    "\n",
-                )
-            try:
-                self.y_pred = self.classifier.predict(self.X_val)
-            except Exception as error:
-                print(
-                    classifiers[self.predictor],
-                    "Data Prediction Failed with error: ",
-                    error,
-                    "\n",
-                )
-            if self.predictor == "ann":
-                self.y_pred = (self.y_pred > 0.5).astype("int32")
+                    self.parameters, self.classifier = classificationPredictor(
+                        self.predictor,
+                        self.params,
+                        self.X_train,
+                        self.X_val,
+                        self.y_train,
+                        self.y_val,
+                        self.epochs,
+                        self.hidden_layers,
+                        self.input_activation,
+                        self.output_activation,
+                        self.loss,
+                        self.batch_size,
+                        self.metrics,
+                        self.validation_split,
+                        self.optimizer,
+                        self.output_units,
+                        self.input_units,
+                        self.tune_mode,
+                        all_mode=True,
+                        verbose=self.verbose,
+                    )
+                elif self.predictor == "ann":
+                    (
+                        self.parameters,
+                        self.classifier,
+                        self.classifier_wrap,
+                    ) = classificationPredictor(
+                        self.predictor,
+                        self.params,
+                        self.X_train,
+                        self.X_val,
+                        self.y_train,
+                        self.y_val,
+                        self.epochs,
+                        self.hidden_layers,
+                        self.input_activation,
+                        self.output_activation,
+                        self.loss,
+                        self.batch_size,
+                        self.metrics,
+                        self.validation_split,
+                        self.optimizer,
+                        self.output_units,
+                        self.input_units,
+                        self.tune_mode,
+                        self.dropout_rate,
+                        all_mode=True,
+                        verbose=self.verbose,
+                    )
+                try:
 
-            # Accuracy ---------------------------------------------------------------------
-            try:
-                self.accuracy = accuracy_score(self.y_val, self.y_pred)
-                self.acc.append(self.accuracy * 100)
-            except Exception as error:
-                print(
-                    classifiers[self.predictor],
-                    "Evaluation Failed with error: ",
-                    error,
-                    "\n",
-                )
+                    if not self.predictor == "ann":
+                        self.classifier.fit(self.X_train, self.y_train)
+                except Exception as error:
+                    print(
+                        classifiers[self.predictor],
+                        "Model Train Failed with error: ",
+                        error,
+                        "\n",
+                    )
+                try:
+                    self.y_pred = self.classifier.predict(self.X_val)
+                except Exception as error:
+                    print(
+                        classifiers[self.predictor],
+                        "Data Prediction Failed with error: ",
+                        error,
+                        "\n",
+                    )
+                if self.predictor == "ann":
+                    self.y_pred = (self.y_pred > 0.5).astype("int32")
 
-            # K-Fold ---------------------------------------------------------------------
-            if self.predictor == "ann":
-                self.classifier_name, self.kfold_accuracy = kfold(
-                    self.classifier_wrap,
-                    self.predictor,
-                    self.X_train,
-                    self.y_train,
-                    self.cv_folds,
-                    all_mode=True,
-                )
-            else:
-                self.classifier_name, self.kfold_accuracy = kfold(
-                    self.classifier,
-                    self.predictor,
-                    self.X_train,
-                    self.y_train,
-                    self.cv_folds,
-                    all_mode=True,
-                )
-            self.kfoldacc.append(self.kfold_accuracy)
-            # GridSearch ---------------------------------------------------------------------
-            if not self.predictor == "nb" and self.tune:
-                self.__tuner(all_mode=True)
-            if self.predictor == "nb":
-                self.best_params = ""
-                self.best_accuracy = self.kfold_accuracy
+                # Accuracy ---------------------------------------------------------------------
+                try:
+                    self.accuracy = accuracy_score(self.y_val, self.y_pred)
+                    self.acc.append(self.accuracy * 100)
+                except Exception as error:
+                    print(
+                        classifiers[self.predictor],
+                        "Evaluation Failed with error: ",
+                        error,
+                        "\n",
+                    )
+
+                # K-Fold ---------------------------------------------------------------------
+                if self.predictor == "ann":
+                    self.classifier_name, self.kfold_accuracy = kfold(
+                        self.classifier_wrap,
+                        self.predictor,
+                        self.X_train,
+                        self.y_train,
+                        self.cv_folds,
+                        all_mode=True,
+                    )
+                else:
+                    self.classifier_name, self.kfold_accuracy = kfold(
+                        self.classifier,
+                        self.predictor,
+                        self.X_train,
+                        self.y_train,
+                        self.cv_folds,
+                        all_mode=True,
+                    )
+                self.kfoldacc.append(self.kfold_accuracy)
+                self.classifier_model.append(self.classifier)
+                # GridSearch ---------------------------------------------------------------------
+                if not self.predictor == "nb" and self.tune:
+                    self.__tuner(all_mode=True)
+                if self.predictor == "nb":
+                    self.best_params = ""
+                    self.best_accuracy = self.kfold_accuracy
         self.result_df["Accuracy"] = self.acc
         self.result_df["KFold Accuracy"] = self.kfoldacc
+        self.result_df["Model"] = self.classifier_model
         if self.tune:
             self.result_df["Best Parameters"] = self.bestparams
             self.result_df["Best Accuracy"] = self.bestacc
 
-        display(self.result_df)
+        self.best_classifier = Best(
+            self.result_df.loc[self.result_df["KFold Accuracy"].idxmax()], self.tune
+        )
+        self.best_classifier_path, self.scaler_path = self.save(
+            best=True, model=self.best_classifier.model, scaler=self.sc
+        )
+        print("Training All Classifiers Done [", u"\u2713", "]\n")
+        print(
+            "Saved Best Model to {} and its scaler to {}".format(
+                self.best_classifier_path, self.scaler_path
+            ),
+            "\n",
+        )
+        display(self.result_df.iloc[:, :-1])
         print("Complete [", u"\u2713", "]\n")
         self.end = time.time()
         print("Time Elapsed : ", self.end - self.start, "seconds \n")
         return
 
-    def _tuner(self, all_mode=False):
+    def __tuner(self, all_mode=False):
         if not all_mode:
             print(
                 "Applying Grid Search Cross validation on Mode {} [*]".format(
@@ -583,15 +631,112 @@ class Classification:
         if not self.pred_mode == "all":
             X_test = np.array(X_test)
             if X_test.ndim == 1:
-                print(
-                    """Input Array has only 1-Dimension but 2-Dimension was expected. 
-                    Reshaping it to 2-Dimension"""
-                )
                 X_test = X_test.reshape(1, -1)
 
-            y_test = self.regressor.predict(self.sc.transform(X_test))
+            y_test = self.classifier.predict(self.sc.transform(X_test))
 
             return y_test
         if self.pred_mode == "all":
-            print("[Error] This method is only applicable on single predictor")
-            return
+            raise TypeError("Predict is only applicable on single predictor")
+
+    def save(self, path=None, **kwargs):
+        """
+        Saves the model and its scaler to a file provided with a path.
+        If no path is provided will create a directory named
+        lucifer_ml_info/models/ and lucifer_ml_info/scaler/ in current working directory
+        Args:
+            path ([list]): [List containing path to save the model and scaler.]
+                Example: path = ["model.pkl", "scaler.pkl"]
+
+        Returns:
+            Path to the saved model and its scaler.
+        """
+        if not type(path) == list and path != None:
+            raise TypeError("Path must be a list")
+
+        if not self.predictor == "all":
+            dir_path_model = path[0] if path else "lucifer_ml_info/models/classifier/"
+            dir_path_scaler = path[1] if path else "lucifer_ml_info/scalers/classifier/"
+            if kwargs.get("best"):
+                dir_path_model = "lucifer_ml_info/best/classifier/models/"
+                dir_path_scaler = "lucifer_ml_info/best/classifier/scalers/"
+            os.makedirs(dir_path_model, exist_ok=True)
+            os.makedirs(dir_path_scaler, exist_ok=True)
+            timestamp = str(int(time.time()))
+            path_model = (
+                dir_path_model
+                + classifiers[self.predictor].replace(" ", "_")
+                + "_"
+                + timestamp
+                + ".pkl"
+            )
+            path_scaler = (
+                dir_path_scaler
+                + classifiers[self.predictor].replace(" ", "_")
+                + "_"
+                + "Scaler"
+                + "_"
+                + timestamp
+                + ".pkl"
+            )
+            if (
+                not kwargs.get("model")
+                and not kwargs.get("best")
+                and not kwargs.get("scaler")
+            ):
+                dump(self.classifier, open(path_model, "wb"))
+                dump(self.sc, open(path_scaler, "wb"))
+            else:
+                dump(kwargs.get("model"), open(path_model, "wb"))
+                dump(kwargs.get("scaler"), open(path_scaler, "wb"))
+            if not kwargs.get("best"):
+                print(
+                    "Model Saved at {} and Scaler at {}".format(path_model, path_scaler)
+                )
+            return path_model, path_scaler
+        else:
+            raise Exception(
+                "[Error] This method is only applicable on single predictor"
+            )
+
+    def __load(self, path=None):
+        """
+        Loads model and scaler from the specified path
+        Args:
+            path ([list]): [List containing path to load the model and scaler.]
+                Example: path = ["model.pkl", "scaler.pkl"]
+
+        Returns:
+            [Model] : [Loaded model]
+            [Scaler] : [Loaded scaler]
+        """
+
+        model_path = path[0] if path[0] else None
+        scaler_path = path[1] if path[1] else None
+        if not ".pkl" in model_path and not model_path == None:
+            raise TypeError(
+                "[Error] Model Filetype not supported. Please use .pkl type "
+            )
+        if not ".pkl" in scaler_path and not scaler_path == None:
+            raise TypeError(
+                "[Error] Scaler Filetype not supported. Please use .pkl type "
+            )
+        if model_path != None and scaler_path != None:
+            model = load(open(model_path, "rb"))
+            scaler = load(open(scaler_path, "rb"))
+            print(
+                "[Info] Model and Scaler Loaded from {} and {}".format(
+                    model_path, scaler_path
+                )
+            )
+            return model, scaler
+        elif model_path != None and scaler_path == None:
+            model = load(open(model_path, "rb"))
+            print("[Info] Model Loaded from {}".format(model_path))
+            return model
+        elif model_path == None and scaler_path != None:
+            scaler = load(open(scaler_path, "rb"))
+            print("[Info] Scaler Loaded from {}".format(scaler_path))
+            return scaler
+        else:
+            raise ValueError("No path specified.Please provide actual path\n")
